#!/usr/bin/env python
"""The NegMAS universal command line tool"""
from __future__ import annotations

import json
import os
import pathlib
import sys
import urllib.request
from functools import partial
from pathlib import Path
from pprint import pprint
from time import perf_counter

import click
import click_config_file
import yaml
from tabulate import tabulate

import negmas
from negmas.genius.common import DEFAULT_JAVA_PORT
from negmas.helpers import humanize_time, unique_name
from negmas.helpers.inout import load
from negmas.tournaments import (
    combine_tournament_results,
    combine_tournament_stats,
    combine_tournaments,
    create_tournament,
    evaluate_tournament,
    run_tournament,
)

try:
    from .vendor.quick.quick import gui_option
except:

    def gui_option(x):
        return x


try:
    # disable a warning in yaml 1b1 version
    yaml.warnings({"YAMLLoadWarning": False})
except:
    pass

n_completed = 0
n_total = 0

GENIUS_JAR_NAME = "geniusbridge.jar"

DEFAULT_NEGOTIATOR = "negmas.sao.AspirationNegotiator"


def default_log_path():
    """Default location for all logs"""

    return Path.home() / "negmas" / "logs"


def default_tournament_path():
    """The default path to store tournament run info"""

    return default_log_path() / "tournaments"


def print_progress(_, i, n) -> None:
    """Prints the progress of a tournament"""
    global n_completed, n_total
    n_completed = i + 1
    n_total = n
    print(
        f"{n_completed:04} of {n:04} worlds completed ({n_completed / n:0.2%})",
        flush=True,
    )


def print_world_progress(world) -> None:
    """Prints the progress of a world"""
    step = world.current_step + 1
    s = (
        f"World# {n_completed:04}: {step:04}  of {world.n_steps:04} "
        f"steps completed ({step / world.n_steps:0.2f}) "
    )

    if n_total > 0:
        s += f"TOTAL: ({n_completed + step / world.n_steps / n_total:0.2f})"
    print(s, flush=True)


click.option = partial(click.option, show_default=True)


@gui_option
@click.group()
def cli():
    pass


@cli.group(chain=True, invoke_without_command=True)
@click.pass_context
@click.option(
    "--ignore-warnings/--show-warnings",
    default=False,
    help="Ignore/show runtime warnings",
)
def tournament(ctx, ignore_warnings):
    if ignore_warnings:
        import warnings

        warnings.filterwarnings("ignore")
    ctx.obj = {}


@tournament.command(help="Creates a tournament")
@click.option(
    "--name",
    "-n",
    default="random",
    help='The name of the tournament. The special value "random" will result in a random name',
)
@click.option(
    "--timeout",
    "-t",
    default=0,
    type=int,
    help="Timeout the whole tournament after the given number of seconds (0 for infinite)",
)
@click.option(
    "--configs",
    default=5,
    type=int,
    help="Number of unique configurations to generate.",
)
@click.option("--runs", default=2, help="Number of runs for each configuration")
@click.option(
    "--max-runs",
    default=-1,
    type=int,
    help="Maximum total number of runs. Zero or negative numbers mean no limit",
)
@click.option(
    "--steps-min",
    default=50,
    type=int,
    help="Minimum number of steps (only used if --steps was not passed",
)
@click.option(
    "--steps-max",
    default=100,
    type=int,
    help="Maximum number of steps (only used if --steps was not passed",
)
@click.option(
    "--agents",
    default=3,
    type=int,
    help="Number of agents per competitor",
)
@click.option(
    "--competitors",
    help="A semicolon (;) separated list of agent types to use for the competition.",
)
@click.option(
    "--non-competitors",
    help="A semicolon (;) separated list of agent types to exist in the worlds as non-competitors "
    "(their scores will not be calculated).",
)
@click.option(
    "--log",
    "-l",
    type=click.Path(dir_okay=True, file_okay=False),
    default=default_tournament_path(),
    help="Default location to save logs (A folder will be created under it)",
)
@click.option(
    "--world-config",
    type=click.Path(dir_okay=False, file_okay=True),
    default=tuple(),
    multiple=True,
    help="A file to load extra configuration parameters for world simulations from.",
)
@click.option(
    "--verbosity",
    default=1,
    type=int,
    help="verbosity level (from 0 == silent to 1 == world progress)",
)
@click.option(
    "--log-ufuns/--no-ufun-logs",
    default=False,
    help="Log ufuns into their own CSV file. Only effective if --debug is given",
)
@click.option(
    "--log-negs/--no-neg-logs",
    default=False,
    help="Log all negotiations. Only effective if --debug is given",
)
@click.option(
    "--compact/--debug",
    default=True,
    help="If True, effort is exerted to reduce the memory footprint which"
    "includes reducing logs dramatically.",
)
@click.option(
    "--raise-exceptions/--ignore-exceptions",
    default=True,
    help="Whether to ignore agent exceptions",
)
@click.option(
    "--path",
    default="",
    help="A path to be added to PYTHONPATH in which all competitors are stored. You can path a : separated list of "
    "paths on linux/mac and a ; separated list in windows",
)
@click.option(
    "--config-generator",
    default="",
    help="The full path to a configuration generator function that is used to generate"
    " all configs for the "
    "tournament. MUST be specified",
)
@click.option(
    "--world-generator",
    default="",
    help="The full path to a world generator function that is used to generate all worlds (given the assigned "
    "configs for the tournament. MUST be specified",
)
@click.option(
    "--assigner",
    default="",
    help="The full path to an assigner function that assigns competitors to different configurations",
)
@click.option("--scorer", default="", help="The full path to a scoring function")
@click.option(
    "--cw",
    default=None,
    type=int,
    help="Number of competitors to run at every world simulation. It must "
    "either be left at default or be a number > 1 and < the number "
    "of competitors passed using --competitors",
)
@click_config_file.configuration_option()
@click.pass_context
def create(
    ctx,
    name,
    timeout,
    log,
    verbosity,
    reveal_names,
    runs,
    configs,
    max_runs,
    competitors,
    world_config,
    non_competitors,
    compact,
    agents,
    log_ufuns,
    log_negs,
    raise_exceptions,
    steps_min,
    steps_max,
    path,
    cw,
    world_generator,
    config_generator,
    assigner,
    scorer,
):
    if len(config_generator is None or config_generator.strip()) == 0:
        print(
            "ERROR: You did not specify a config generator. Use --config-generator to specify one and see the "
            "documentation of the create_tournament method in negmas.situated for details about it."
            "\nThe following must be explicitly specified to create a tournament: a world-generator, "
            "an assigner, a scorer, and a config-generator."
        )

        return -4

    if len(world_generator is None or world_generator.strip()) == 0:
        print(
            "ERROR: You did not specify a world generator. Use --world-generator to specify one and see the "
            "documentation of the create_tournament method in negmas.situated for details about it."
            "\nThe following must be explicitly specified to create a tournament: a world-generator, "
            "an assigner, a scorer, and a config-generator."
        )

        return -3

    if len(assigner is None or assigner.strip()) == 0:
        print(
            "ERROR: You did not specify an assigner. Use --assigner to specify one and see the documentation"
            " of the create_tournament method in negmas.situated for details about it."
            "\nThe following must be explicitly specified to create a tournament: a world-generator, "
            "an assigner, a scorer, and a config-generator."
        )

        return -2

    if len(scorer is None or scorer.strip()) == 0:
        print(
            "ERROR: You did not specify a scorer. Use --scorer to specify one and see the documentation of the "
            "create_tournament method in negmas.situated for details about it."
            "\nThe following must be explicitly specified to create a tournament: a world-generator, "
            "an assigner, a scorer, and a config-generator."
        )

        return -1

    if len(path) > 0:
        sys.path.append(path)
    kwargs = {}

    if world_config is not None and len(world_config) > 0:
        for wc in world_config:
            kwargs.update(load(wc))
    warning_n_runs = 2000

    if timeout <= 0:
        timeout = None

    if name == "random":
        name = unique_name(base="", rand_digits=0)
    ctx.obj["tournament_name"] = name

    if max_runs <= 0:
        max_runs = None

    if compact:
        log_ufuns = False

    if not compact:
        if not reveal_names:
            print(
                "You are running the tournament with --debug. Will reveal "
                "agent types in their names"
            )
        reveal_names = True
        verbosity = max(1, verbosity)

    worlds_per_config = (
        None if max_runs is None else int(round(max_runs / (configs * runs)))
    )

    all_competitors = competitors.split(";")
    all_competitors_params = [dict() for _ in range(len(all_competitors))]

    permutation_size = 1
    recommended = runs * configs * permutation_size

    if worlds_per_config is not None and worlds_per_config < 1:
        print(
            f"You need at least {(configs * runs)} runs even with a single permutation of managers."
            f".\n\nSet --max-runs to at least {(configs * runs)} (Recommended {recommended})"
        )

        return

    if max_runs is not None and max_runs < recommended:
        print(
            f"You are running {max_runs} worlds only but it is recommended to set {max_runs} to at least "
            f"{recommended}. Will continue"
        )

    steps = (steps_min, steps_max)

    if worlds_per_config is None:
        len(all_competitors)
        n_worlds = permutation_size * runs * configs

        if n_worlds > warning_n_runs:
            print(
                f"You are running the maximum possible number of permutations for each configuration. This is roughly"
                f" {n_worlds} simulations (each for {steps} steps). That will take a VERY long time."
                f"\n\nYou can reduce the number of simulations by setting --configs>=1 (currently {configs}) or "
                f"--runs>= 1 (currently {runs}) to a lower value. "
                f"\nFinally, you can limit the maximum number of worlds to run by setting --max-runs=integer."
            )
            # if (
            #     not input(f"Are you sure you want to run {n_worlds} simulations?")
            #     .lower()
            #     .startswith("y")
            # ):
            #     exit(0)
            max_runs = int(
                input(
                    f"Input the maximum number of simulations to run. Zero to run all of the {n_worlds} "
                    f"simulations. ^C or a negative number to exit [0 : {n_worlds}]:"
                )
            )

            if max_runs == 0:
                max_runs = None

            if max_runs is not None and max_runs < 0:
                exit(0)
            worlds_per_config = (
                None if max_runs is None else int(round(max_runs / (configs * runs)))
            )

    non_competitor_params = None

    if len(non_competitors) < 1:
        non_competitors = None
    else:
        non_competitors = non_competitors.split(";")

    print(f"Tournament will be run between {len(all_competitors)} agents: ")
    pprint(all_competitors)
    print("Non-competitors are: ")
    pprint(non_competitors)
    results = create_tournament(
        competitors=all_competitors,
        competitor_params=all_competitors_params,
        non_competitors=non_competitors,
        non_competitor_params=non_competitor_params,
        agent_names_reveal_type=reveal_names,
        n_competitors_per_world=cw,
        n_configs=configs,
        n_runs_per_world=runs,
        max_worlds_per_config=worlds_per_config,
        base_tournament_path=log,
        total_timeout=timeout,
        name=name,
        verbose=verbosity > 0,
        n_agents_per_competitor=agents,
        world_generator=world_generator,
        config_generator=config_generator,
        config_assigner=assigner,
        score_calculator=scorer,
        compact=compact,
        n_steps=steps,
        log_ufuns=log_ufuns,
        log_negotiations=log_negs,
        ignore_agent_exceptions=not raise_exceptions,
        ignore_contract_execution_exceptions=not raise_exceptions,
        **kwargs,
    )
    results = Path(results)
    ctx.obj["tournament_name"] = results.name
    ctx.obj["tournament_log_folder"] = log
    ctx.obj["compact"] = compact
    print(f"Saved all configs to {str(results)}\nTournament name is {results.name}")


def display_results(results, metric, significance):
    if metric == "truncated_mean":
        print(
            tabulate(
                results.total_scores.sort_values(by="score", ascending=False),
                headers="keys",
                tablefmt="psql",
            )
        )
        print(
            tabulate(
                results.score_stats.sort_values(by="median", ascending=False),
                headers="keys",
                tablefmt="psql",
            )
        )
    else:
        viewmetric = ["50%" if metric == "median" else metric]
        print(
            tabulate(
                results.score_stats.sort_values(by=viewmetric, ascending=False),
                headers="keys",
                tablefmt="psql",
            )
        )

    if significance:
        if metric in ("mean", "sum", "tuncated_mean"):
            print(tabulate(results.ttest, headers="keys", tablefmt="psql"))
        else:
            print(tabulate(results.kstest, headers="keys", tablefmt="psql"))

    try:
        agg_stats = results.agg_stats.loc[
            :,
            [
                "n_negotiations_sum",
                "n_contracts_concluded_sum",
                "n_contracts_signed_sum",
                "n_contracts_executed_sum",
                "activity_level_sum",
            ],
        ]
        agg_stats.columns = [
            "negotiated",
            "concluded",
            "signed",
            "executed",
            "business",
        ]
        print(tabulate(agg_stats.describe(), headers="keys", tablefmt="psql"))
    except:
        pass


@tournament.command(help="Runs/continues a tournament")
@click.option(
    "--name",
    "-n",
    default="",
    help="The name of the tournament. When invoked after create, there is no need to pass it",
)
@click.option(
    "--log",
    "-l",
    type=click.Path(dir_okay=True, file_okay=False),
    default=default_tournament_path(),
    help="Default location to save logs",
)
@click.option(
    "--verbosity",
    default=1,
    type=int,
    help="verbosity level (from 0 == silent to 1 == world progress)",
)
@click.option(
    "--parallel/--serial",
    default=True,
    help="Run a parallel/serial tournament on a single machine",
)
@click.option(
    "--distributed/--single-machine",
    default=False,
    help="Run a distributed tournament using dask",
)
@click.option(
    "--ip",
    default="127.0.0.1",
    help="The IP address for a dask scheduler to run the distributed tournament."
    " Effective only if --distributed",
)
@click.option(
    "--port",
    default=8786,
    type=int,
    help="The IP port number a dask scheduler to run the distributed tournament."
    " Effective only if --distributed",
)
@click.option(
    "--compact/--debug",
    default=True,
    help="If True, effort is exerted to reduce the memory footprint which"
    "includes reducing logs dramatically.",
)
@click.option(
    "--path",
    default="",
    help="A path to be added to PYTHONPATH in which all competitors are stored. You can path a : separated list of "
    "paths on linux/mac and a ; separated list in windows",
)
@click.option(
    "--metric",
    default="truncated_mean",
    type=str,
    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
)
@click.option(
    "--significance/--no-significance",
    default=False,
    help="Whether to show significance table",
)
@click.option(
    "--eval/--no-eval",
    default=False,
    help="Whether evaluate and show results after the tournament is run",
)
@click_config_file.configuration_option()
@click.pass_context
def run(
    ctx,
    name,
    verbosity,
    parallel,
    distributed,
    ip,
    port,
    compact,
    path,
    log,
    metric,
    significance,
    eval,
):
    if len(name) == 0:
        name = ctx.obj.get("tournament_name", "")

    if len(name) == 0:
        print(
            "Name is not given to run command and was not stored during a create command call"
        )
        exit(1)

    if len(path) > 0:
        sys.path.append(path)
    saved_log_folder = ctx.obj.get("tournament_log_folder", None)

    if saved_log_folder is not None:
        log = saved_log_folder
    parallelism = "distributed" if distributed else "parallel" if parallel else "serial"
    prog_callback = print_world_progress if verbosity > 1 and not distributed else None
    tpath = str(pathlib.Path(log) / name)
    start = perf_counter()
    run_tournament(
        tournament_path=tpath,
        verbose=verbosity > 0,
        compact=compact,
        world_progress_callback=prog_callback,
        parallelism=parallelism,
        scheduler_ip=ip,
        scheduler_port=port,
        print_exceptions=verbosity > 1,
    )
    end_time = humanize_time(perf_counter() - start)
    if eval:
        results = evaluate_tournament(
            tournament_path=tpath, verbose=verbosity > 0, metric=metric, compile=True
        )
        display_results(results, metric, significance)
    print(f"Finished in {end_time}")


@tournament.command(help="Evaluates a tournament and returns the results")
@click.argument(
    "path",
    type=click.Path(dir_okay=True, file_okay=False),
)
@click.option(
    "--metric",
    default="truncated_mean",
    type=str,
    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
)
@click.option(
    "--significance/--no-significance",
    default=False,
    help="Whether to show significance table",
)
@click.option(
    "--compile/--show",
    default=True,
    help="Whether to recompile results from individual world runs or just show the already-compiled results",
)
@click.option(
    "--verbose/--silent",
    default=True,
    help="Whether to be verbose",
)
@click_config_file.configuration_option()
@click.pass_context
def eval(ctx, path, metric, significance, compile, verbose):
    results = evaluate_tournament(
        tournament_path=path, metric=metric, compile=compile, verbose=verbose
    )
    display_results(results, metric, significance)


@tournament.command(
    help="Finds winners of a tournament or a set of tournaments sharing a root"
)
@click.option(
    "--name",
    "-n",
    default="",
    help="The name of the tournament. When invoked after create, there is no need to pass it",
)
@click.option(
    "--log",
    "-l",
    type=click.Path(dir_okay=True, file_okay=False),
    default=default_tournament_path(),
    help="Default location to save logs",
)
@click.option(
    "--recursive/--no-recursive",
    default=True,
    help="Whether to recursively look for tournament results. --name should not be given if --recursive",
)
@click.option(
    "--metric",
    default="truncated_mean",
    type=str,
    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
)
@click.option(
    "--verbose/--silent",
    default=True,
    help="Whether to be verbose",
)
@click.option(
    "--significance/--no-significance",
    default=False,
    help="Whether to show significance table",
)
@click.option(
    "--compile/--show",
    default=True,
    help="Whether to recompile results from individual world runs or just show the already-compiled results",
)
@click_config_file.configuration_option()
@click.pass_context
def winners(ctx, name, log, recursive, metric, significance, compile, verbose):
    if len(name) == 0:
        if not recursive:
            name = ctx.obj.get("tournament_name", "")
        else:
            name = None

    if (name is None or len(name) == 0) and not recursive:
        print(
            "Name is not given to run command and was not stored during a create command call"
        )
        exit(1)
    saved_log_folder = ctx.obj.get("tournament_log_folder", None)

    if saved_log_folder is not None:
        log = saved_log_folder

    if name is not None and len(name) > 0:
        tpath = str(pathlib.Path(log) / name)
    else:
        tpath = str(pathlib.Path(log))
    results = evaluate_tournament(
        tournament_path=tpath,
        verbose=verbose,
        recursive=recursive,
        metric=metric,
        compile=compile,
    )
    display_results(results, metric, significance)


def _path(path) -> Path:
    """Creates an absolute path from given path which can be a string"""

    if isinstance(path, Path):
        return path.absolute()
    path.replace("/", os.sep)

    if isinstance(path, str):
        if path.startswith("~"):
            path = Path.home() / (os.sep.join(path.split(os.sep)[1:]))

    return Path(path).absolute()


@tournament.command(help="Combine multiple tournaments at the given base path(s)")
@click.argument(
    "path",
    type=click.Path(dir_okay=True, file_okay=False),
    nargs=-1,
)
@click.option(
    "--verbose/--silent",
    default=True,
    help="Whether to be verbose",
)
@click.option(
    "--dest",
    "-d",
    type=click.Path(dir_okay=True, file_okay=False),
    help="The location to save the results",
    default=None,
)
@click_config_file.configuration_option()
def combine(path, dest, verbose):
    if dest is None:
        print("Must specify the destination using --dest/-d option")
        return
    tpath = [_path(_) for _ in path]

    if len(tpath) < 1:
        print("No paths are given to combine")
    combine_tournaments(sources=tpath, dest=dest, verbose=verbose)


@tournament.command(help="Combine results from multiple tournaments")
@click.argument("path", type=click.Path(dir_okay=True, file_okay=False), nargs=-1)
@click.option(
    "--dest",
    "-d",
    type=click.Path(dir_okay=True, file_okay=False),
    help="The location to save the results",
    default=None,
)
@click.option(
    "--metric",
    default="truncated_mean",
    type=str,
    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
)
@click.option(
    "--significance/--no-significance",
    default=False,
    help="Whether to show significance table",
)
@click.option(
    "--verbose/--silent",
    default=True,
    help="Whether to be verbose",
)
@click.option(
    "--compile/--show",
    default=True,
    help="Whether to recompile results from individual world runs or just show the already-compiled results",
)
@click_config_file.configuration_option()
def combine_results(path, dest, metric, significance, compile, verbose):
    tpath = [_path(_) for _ in path]

    if len(tpath) < 1:
        print("No paths are given to combine")
    scores = combine_tournament_results(sources=tpath, dest=None, verbose=verbose)
    stats = combine_tournament_stats(sources=tpath, dest=None, verbose=verbose)
    results = evaluate_tournament(
        dest, scores, stats, verbose=verbose, metric=metric, compile=compile
    )
    display_results(results, metric, significance)


@cli.command(help="Start the bridge to genius (to use GeniusNegotiator)")
@click.option(
    "--path",
    "-p",
    default="auto",
    help='Path to geniusbridge.jar with embedded NegLoader. Use "auto" to '
    "read the path from ~/negmas/config.json"
    "\n\tConfig key is genius_bridge_jar"
    "\nYou can download this jar from: "
    "http://www.yasserm.com/scml/geniusbridge.jar",
)
@click.option(
    "--port",
    "-r",
    default=DEFAULT_JAVA_PORT,
    help="Port to run the NegLoader on. Pass 0 for the default value",
)
@click.option(
    "--debug/--silent",
    default=False,
    help="Run the bridge in debug mode if --debug else silently",
)
@click.option(
    "--timeout",
    default=0,
    type=float,
    help="The timeout to pass. Zero or negative numbers to disable and use the bridge's global timeout.",
)
def genius(path, port, debug, timeout):
    if port and negmas.genius_bridge_is_running(port):
        print(f"Genius Bridge is already running on port {port} ... exiting")
        sys.exit()
    negmas.init_genius_bridge(
        path=path if path != "auto" else None,
        port=port,
        debug=debug,
        timeout=timeout,
    )
    while True:
        pass


def download_and_set(key, url, file_name):
    """
    Downloads a file and sets the corresponding key in ~/negmas/config.json

    Args:
        key: Key name in config.json
        url: URL to download from
        file_name: file name

    Returns:

    """
    config_path = Path.home() / "negmas"
    jar_path = config_path / "files"
    jar_path.mkdir(parents=True, exist_ok=True)
    jar_path = jar_path / file_name
    urllib.request.urlretrieve(url, jar_path)
    config = {}
    config_file = config_path / "config.json"

    if config_file.exists():
        with open(config_file) as f:
            config = json.load(f)
    config[key] = str(jar_path)
    with open(config_file, "w") as f:
        json.dump(config, fp=f, sort_keys=True, indent=4)


@cli.command(help="Downloads the genius bridge and updates your settings")
def genius_setup():
    url = f"http://www.yasserm.com/scml/{GENIUS_JAR_NAME}"
    print(f"Downloading: {url}", end="", flush=True)
    download_and_set(key="genius_bridge_jar", url=url, file_name=GENIUS_JAR_NAME)
    print(" done successfully")


@cli.command(help="Prints NegMAS version")
def version():
    print(negmas.__version__)


if __name__ == "__main__":
    cli()



from __future__ import annotations

from datetime import timedelta
from pathlib import Path, PosixPath

import hypothesis.strategies as st
from hypothesis import HealthCheck, example, given, settings
from pytest import mark

from negmas import AspirationNegotiator, MappingUtilityFunction, SAOMechanism
from negmas.checkpoints import CheckpointRunner
from negmas.helpers import unique_name


def checkpoint_every(args):
    pass


# @settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
# @given(
#     single_checkpoint=st.booleans(),
#     checkpoint_every=st.integers(1, 4),
#     exist_ok=st.booleans(),
# )
# def test_can_run_from_checkpoint1(
#     tmp_path, single_checkpoint, checkpoint_every, exist_ok
# ):
#     import shutil
#
#     new_folder: Path = tmp_path / unique_name("empty", sep="")
#     second_folder: Path = tmp_path / unique_name("second", sep="")
#     new_folder.mkdir(parents=True, exist_ok=True)
#     shutil.rmtree(new_folder)
#     new_folder.mkdir(parents=True, exist_ok=True)
#     filename = "mechanism"
#     second_folder.mkdir(parents=True, exist_ok=True)
#
#     n_outcomes, n_negotiators = 5, 3
#     n_steps = 50
#     mechanism = SAOMechanism(
#         outcomes=n_outcomes,
#         n_steps=n_steps,
#         offering_is_accepting=True,
#         avoid_ultimatum=False,
#         checkpoint_every=checkpoint_every,
#         checkpoint_folder=new_folder,
#         checkpoint_filename=filename,
#         extra_checkpoint_info=None,
#         exist_ok=exist_ok,
#         single_checkpoint=single_checkpoint,
#     )
#     ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
#     for i in range(n_negotiators):
#         mechanism.add(AspirationNegotiator(name=f"agent{i}"), ufun=ufuns[i])
#
#     mechanism.run()
#     files = list(new_folder.glob("*"))
#     if 0 < checkpoint_every <= n_steps:
#         if single_checkpoint:
#             assert len(list(new_folder.glob("*"))) == 2
#         else:
#             assert len(list(new_folder.glob("*"))) >= 2 * (
#                 max(1, mechanism.state.step // checkpoint_every)
#             )
#     elif checkpoint_every > n_steps:
#         assert len(list(new_folder.glob("*"))) == 2
#     else:
#         assert len(list(new_folder.glob("*"))) == 0
#
#     runner = CheckpointRunner(folder=new_folder)
#
#     assert len(runner.steps) * 2 == len(files)
#     assert runner.current_step == -1
#     assert runner.loaded_object is None
#
#     runner.step()
#
#     assert runner.current_step == (0 if not single_checkpoint else runner.steps[-1])
#     assert isinstance(runner.loaded_object, SAOMechanism)
#     assert runner.loaded_object.state.step == runner.current_step
#
#     runner.reset()
#     assert len(runner.steps) * 2 == len(files)
#     assert runner.current_step == -1
#     assert runner.loaded_object is None
#
#     runner.goto(runner.last_step, exact=True)
#     assert isinstance(runner.loaded_object, SAOMechanism)
#     assert runner.loaded_object.state.step == runner.current_step
#
#     runner.goto(runner.next_step, exact=True)
#     assert isinstance(runner.loaded_object, SAOMechanism)
#     assert runner.loaded_object.state.step == runner.current_step
#
#     runner.goto(runner.previous_step, exact=True)
#     assert isinstance(runner.loaded_object, SAOMechanism)
#     assert runner.loaded_object.state.step == runner.current_step
#
#     runner.goto(runner.first_step, exact=True)
#     assert isinstance(runner.loaded_object, SAOMechanism)
#     assert runner.loaded_object.state.step == runner.current_step
#
#     runner.reset()
#
#     runner.run()
#


@mark.xfail(
    run=False,
    reason="Checkpointing is known to fail with UtilityInverter. As this whole thing will be changed, we may just wait for now",
)
@given(
    checkpoint_every=st.integers(1, 4),
    exist_ok=st.booleans(),
    copy=st.booleans(),
    fork_after_reset=st.booleans(),
)
@settings(
    deadline=20000,
    max_examples=100,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
@example(
    checkpoint_every=1,
    exist_ok=False,
    copy=False,
    fork_after_reset=False,
    tmp_path=PosixPath(
        "/private/var/folders/zr/p1v4wpjn6cq7yz5y1x6wjydh0000gn/T/pytest-of-yasser/pytest-15/test_can_run_from_checkpoint0"
    ),
)
def test_can_run_from_checkpoint(
    tmp_path, checkpoint_every, exist_ok, copy, fork_after_reset
):
    import shutil

    new_folder: Path = tmp_path / unique_name("empty", sep="")
    second_folder: Path = tmp_path / unique_name("second", sep="")
    new_folder.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(new_folder)
    new_folder.mkdir(parents=True, exist_ok=True)
    second_folder.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(second_folder)
    second_folder.mkdir(parents=True, exist_ok=True)

    filename = "mechanism"

    n_outcomes, n_negotiators = 5, 3
    n_steps = 50
    mechanism = SAOMechanism(
        outcomes=n_outcomes,
        n_steps=n_steps,
        offering_is_accepting=True,
        avoid_ultimatum=False,
        checkpoint_every=checkpoint_every,
        checkpoint_folder=new_folder,
        checkpoint_filename=filename,
        extra_checkpoint_info=None,
        exist_ok=exist_ok,
        single_checkpoint=False,
    )
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(
            AspirationNegotiator(name=f"agent{i}", aspiration_type="conceder"),
            preferences=ufuns[i],
        )

    mechanism.run()
    files = list(new_folder.glob("*"))
    if 0 < checkpoint_every <= n_steps:
        assert len(list(new_folder.glob("*"))) >= 2 * (
            max(1, mechanism.state.step // checkpoint_every)
        )
    elif checkpoint_every > n_steps:
        assert len(list(new_folder.glob("*"))) == 2
    else:
        assert len(list(new_folder.glob("*"))) == 0

    runner = CheckpointRunner(folder=new_folder)

    assert len(runner.steps) * 2 == len(files)
    assert runner.current_step == -1
    assert runner.loaded_object is None

    runner.step()

    assert runner.current_step == 0
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.reset()
    assert len(runner.steps) * 2 == len(files)
    assert runner.current_step == -1
    assert runner.loaded_object is None

    runner.goto(runner.last_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.goto(runner.next_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.goto(runner.previous_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.goto(runner.first_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.reset()

    if fork_after_reset:
        m = runner.fork(copy_past_checkpoints=copy, folder=second_folder)
        assert m is None
        return
    runner.step()
    m = runner.fork(copy_past_checkpoints=copy, folder=second_folder)

    if copy:
        step = runner.current_step
        assert len(list(second_folder.glob("*"))) >= 2
        assert len(list(second_folder.glob(f"*{step}.mechanism"))) > 0
    else:
        assert len(list(second_folder.glob("*"))) == 0

    assert isinstance(m, SAOMechanism)
    step = m.current_step

    m.step()
    assert m.current_step == step + 1

    state = m.run()
    assert state.agreement is not None

    runner.reset()
    assert len(runner.steps) * 2 == len(files)
    assert runner.current_step == -1
    assert runner.loaded_object is None

    runner.goto(runner.last_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.goto(runner.next_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.goto(runner.previous_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.goto(runner.first_step, exact=True)
    assert isinstance(runner.loaded_object, SAOMechanism)
    assert runner.loaded_object.state.step == runner.current_step

    runner.reset()

    runner.run()



"""
Acceptance modeling.
"""
from __future__ import annotations

import random
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Collection

import numpy as np

from negmas.common import NegotiatorMechanismInterface
from negmas.outcomes import Outcome
from negmas.sao import ResponseType

if TYPE_CHECKING:
    from negmas.sao import SAONegotiator

__all__ = [
    "AcceptanceModelType",
    "DiscreteAcceptanceModel",
    "AdaptiveDiscreteAcceptanceModel",
    "RandomDiscreteAcceptanceModel",
    "PeekingDiscreteAcceptanceModel",
    "AggregatingDiscreteAcceptanceModel",
    "UncertainOpponentModel",
    "PeekingProbabilisticDiscreteAcceptanceModel",
]


class AcceptanceModelType(Enum):
    """Represents types of acceptance models."""

    ACCEPTANCE_MODEL_RANDOM = -100
    ACCEPTANCE_MODEL_AUTO = -1
    ACCEPTANCE_MODEL_STATIC = 0
    ACCEPTANCE_MODEL_HOMOGENEOUS_CONST = 1
    ACCEPTANCE_MODEL_HOMOGENEOUS = 2
    ACCEPTANCE_MODEL_MONOTONIC = 3
    ACCEPTANCE_MODEL_GENERAL = 4
    ACCEPTANCE_MODEL_BINARY_STATIC = 7
    ACCEPTANCE_MODEL_BINARY_HOMOGENEOUS_CONST = 8
    ACCEPTANCE_MODEL_BINARY_HOMOGENEOUS = 9
    ACCEPTANCE_MODEL_BINARY_MONOTONIC = 10
    ACCEPTANCE_MODEL_BINARY = 11


class DiscreteAcceptanceModel(ABC):
    """"""

    def __init__(self, outcomes: Collection[Outcome]):
        outcomes = list(outcomes)
        self.outcomes = outcomes
        self.indx = dict(zip(outcomes, range(len(outcomes))))

    @abstractmethod
    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        raise NotImplementedError()

    def probability_of_acceptance(self, outcome: Outcome):
        indx = self.indx.get(outcome, None)
        if indx is None:
            return 0.0
        return self.probability_of_acceptance_indx(indx)

    @abstractmethod
    def update_rejected_indx(self, outcome_index: int):
        raise NotImplementedError()

    def update_rejected(self, outcome: Outcome):
        if outcome is None:
            return
        return self.update_rejected_indx(self.indx[outcome])

    @abstractmethod
    def update_offered_indx(self, outcome_index: int):
        raise NotImplementedError()

    def update_offered(self, outcome):
        if outcome is None:
            return
        return self.update_offered_indx(self.indx[outcome])

    def update_accepted(self, outcome):
        return self.update_offered(outcome=outcome)

    def acceptance_probabilities(self) -> np.ndarray:
        return np.array(
            [self.probability_of_acceptance_indx(_) for _ in range(len(self.outcomes))]
        )


class AdaptiveDiscreteAcceptanceModel(DiscreteAcceptanceModel):
    def __init__(
        self,
        outcomes: Collection[Outcome],
        n_negotiators: int = 2,
        prob: float | list[float] = 0.5,
        end_prob=0.0,
        p_accept_after_reject=0.0,
        p_reject_after_accept=0.0,
        rejection_discount=0.98,
        rejection_delta=0.0,
        not_offering_rejection_ratio=0.75,
    ):
        super().__init__(outcomes=outcomes)
        outcomes = self.outcomes
        if isinstance(prob, list) and len(outcomes) != len(prob):
            raise ValueError(
                f"{len(outcomes)} outcomes but {len(prob)} probabilities. Cannot initialize simple "
                f"opponents model"
            )
        self.n_agents = n_negotiators
        if not isinstance(prob, Collection):
            self.p = np.array([prob for _ in range(len(outcomes))])
        else:
            self.p = np.array(list(prob))
        self.end_prob = end_prob
        self.p_accept_after_reject = p_accept_after_reject
        self.p_accept_after_accept = 1 - p_reject_after_accept
        self.delta = rejection_delta
        self.discount = rejection_discount
        self.first = True
        self.not_offered = set(list(range(len(self.outcomes))))
        self.not_offering_rejection_ratio = not_offering_rejection_ratio
        self.not_offering_discount = self.discount + (
            1.0 - self.not_offering_rejection_ratio
        ) * (1.0 - self.discount)

    @classmethod
    def from_negotiation(
        cls,
        nmi: NegotiatorMechanismInterface,
        prob: float | list = 0.5,
        end_prob=0.0,
        p_accept_after_reject=0.0,
        p_reject_after_accept=0.0,
    ) -> AdaptiveDiscreteAcceptanceModel:
        if not nmi.n_outcomes or nmi.outcomes is None:
            raise ValueError(
                "Cannot initialize this simple opponents model for a negotiation with uncountable outcomes"
            )
        return cls(
            outcomes=nmi.outcomes,
            n_negotiators=nmi.n_negotiators,
            prob=prob,
            end_prob=end_prob,
            p_accept_after_reject=p_accept_after_reject,
            p_reject_after_accept=p_reject_after_accept,
        )

    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        return self.p[outcome_index]

    def acceptance_probabilities(self):
        """Probability of acceptance for all outcomes"""
        return self.p

    def _update(self, p: float, real_rejection: bool) -> float:
        if real_rejection:
            return min(
                self.p_accept_after_reject, min(1.0, (p - self.delta) * self.discount)
            )
        else:
            return max(
                self.p_accept_after_reject,
                min(
                    1.0,
                    (p - self.delta * self.not_offering_rejection_ratio)
                    * self.not_offering_discount,
                ),
            )

    def update_rejected_indx(self, outcome_index: int):
        self.p[outcome_index] = self._update(self.p[outcome_index], real_rejection=True)

    def update_offered_indx(self, outcome_index: int):
        try:
            self.not_offered.remove(outcome_index)
            self.p[outcome_index] = self.p_accept_after_accept
            # for i in self.not_offered:
            #    self.p[i] = self._update(self.p[i], real_rejection=False)
        except KeyError:
            pass


class RandomDiscreteAcceptanceModel(DiscreteAcceptanceModel):
    def __init__(self, outcomes: Collection[Outcome], **kwargs):
        super().__init__(outcomes=outcomes)

    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        return random.random()

    def update_rejected_indx(self, outcome_index: int):
        pass

    def update_offered_indx(self, outcome_index: int):
        pass


class ConstantDiscreteAcceptanceModel(DiscreteAcceptanceModel):
    def __init__(self, outcomes: Collection[Outcome], **kwargs):
        super().__init__(outcomes=outcomes)

    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        return 0.5

    def update_rejected_indx(self, outcome_index: int):
        pass

    def update_offered_indx(self, outcome_index: int):
        pass


class PeekingDiscreteAcceptanceModel(DiscreteAcceptanceModel):
    def __init__(
        self,
        outcomes: Collection[Outcome],
        opponents: SAONegotiator | Collection[SAONegotiator],
    ):
        super().__init__(outcomes=outcomes)
        if not isinstance(opponents, Collection):
            opponents = [opponents]
        self.opponents = opponents

    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        outcome = self.outcomes[outcome_index]
        for opponent in self.opponents:
            if opponent is self:
                continue
            if opponent._nmi is None:
                response = ResponseType.REJECT_OFFER
            else:
                response = opponent.respond_(state=opponent._nmi.state, offer=outcome)  # type: ignore
            if response != ResponseType.ACCEPT_OFFER:
                return 0.0
        return 1.0

    def update_rejected_indx(self, outcome_index: int):
        pass

    def update_offered_indx(self, outcome_index: int):
        pass


class PeekingProbabilisticDiscreteAcceptanceModel(DiscreteAcceptanceModel):
    def __init__(
        self,
        outcomes: Collection[Outcome],
        opponents: SAONegotiator | Collection[SAONegotiator],
    ):
        super().__init__(outcomes=outcomes)
        if not isinstance(opponents, Collection):
            opponents = [opponents]
        self.opponents = opponents

    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        outcome = self.outcomes[outcome_index]
        if outcome is None:
            return 0.0
        prod = 1.0
        for o in self.opponents:
            prod *= o.ufun(outcome)  # type: ignore
        return prod

    def update_rejected_indx(self, outcome_index: int):
        pass

    def update_offered_indx(self, outcome_index: int):
        pass


class AggregatingDiscreteAcceptanceModel(DiscreteAcceptanceModel):
    def __init__(
        self,
        outcomes: Collection[Outcome],
        models: list[DiscreteAcceptanceModel],
        weights: list[float] | None = None,
    ):
        super().__init__(outcomes=outcomes)
        if weights is None:
            weights = [1.0] * len(self.outcomes)
        s = sum(weights)
        weights = [_ / s for _ in weights]
        self.models = models
        self.weights = weights

    def probability_of_acceptance_indx(self, outcome_index: int) -> float:
        p = 0.0
        for model, w in zip(self.models, self.weights):
            p += w * model.probability_of_acceptance_indx(outcome_index=outcome_index)
        return min(1.0, max(0.0, p))

    def update_rejected_indx(self, outcome_index: int):
        for model in self.models:
            model.update_rejected_indx(outcome_index=outcome_index)

    def update_offered_indx(self, outcome_index: int):
        for model in self.models:
            model.update_offered_indx(outcome_index=outcome_index)


class UncertainOpponentModel(AggregatingDiscreteAcceptanceModel):
    """A model for which the uncertainty about the acceptance probability of different negotiators is controllable.

    This is not a realistic model but it can be used to experiment with effects of this uncertainty on different
    negotiation related algorithms (e.g. elicitation algorithms)

    Args:
        outcomes: The list of possible outcomes
        uncertainty (float): The uncertainty level. Zero means no uncertainty and 1.0 means maximum uncertainty
        adaptive (bool): If true then the random part will learn from experience with the opponents otherwise it will not.
        rejection_discount: Only effective if adaptive is True. See `AdaptiveDiscreteAcceptanceModel`
        rejection_delta: Only effective if adaptive is True. See `AdaptiveDiscreteAcceptanceModel`

    """

    def __init__(
        self,
        outcomes: Collection[Outcome],
        opponents: SAONegotiator | Collection[SAONegotiator],
        uncertainty: float = 0.5,
        adaptive: bool = False,
        rejection_discount: float = 0.95,
        rejection_delta: float = 0.0,
        constant_base=True,
        accesses_real_acceptance=False,
    ):
        randomizing_model: DiscreteAcceptanceModel
        peaking_model: DiscreteAcceptanceModel
        if adaptive:
            randomizing_model = AdaptiveDiscreteAcceptanceModel(
                outcomes=outcomes,
                rejection_discount=rejection_discount,
                rejection_delta=rejection_delta,
            )
        elif constant_base:
            randomizing_model = ConstantDiscreteAcceptanceModel(outcomes=outcomes)
        else:
            randomizing_model = RandomDiscreteAcceptanceModel(outcomes=outcomes)
        if accesses_real_acceptance:
            peaking_model = PeekingDiscreteAcceptanceModel(
                opponents=opponents, outcomes=outcomes
            )
        else:
            peaking_model = PeekingProbabilisticDiscreteAcceptanceModel(
                opponents=opponents, outcomes=outcomes
            )
        if uncertainty < 1e-7:
            super().__init__(outcomes=outcomes, models=[peaking_model], weights=[1.0])
        elif uncertainty > 1.0 - 1e-7:
            super().__init__(
                outcomes=outcomes, models=[randomizing_model], weights=[1.0]
            )
        else:
            super().__init__(
                outcomes=outcomes,
                models=[peaking_model, randomizing_model],
                weights=[1.0 - uncertainty, uncertainty],
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Collection

from negmas.common import NegotiatorMechanismInterface
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue
from negmas.preferences import Preferences

from .action import Action
from .common import NegotiationRequestInfo, RunningNegotiationInfo
from .contract import Contract

if TYPE_CHECKING:
    from .agent import Agent
    from .world import World

__all__ = ["AgentWorldInterface"]


class AgentWorldInterface:
    """Agent World Interface class"""

    # def __getstate__(self):
    #     return self._world, self.agent.id
    #
    # def __setstate__(self, state):
    #     self._world, agent_id = state
    #     self.agent = self._world.agents[agent_id]

    def __init__(self, world: World, agent: Agent):
        self._world, self.agent = world, agent

    def execute(
        self, action: Action, callback: Callable[[Action, bool], Any] | None = None
    ) -> bool:
        """Executes an action in the world simulation"""
        return self._world.execute_action(
            action=action, agent=self.agent, callback=callback
        )

    @property
    def state(self) -> Any:
        """Returns the private state of the agent in that world"""
        return self._world.get_private_state(self.agent)

    @property
    def relative_time(self) -> float:
        """Relative time of the simulation going from 0 to 1"""
        return self._world.relative_time

    @property
    def current_step(self) -> int:
        """Current simulation step"""
        return self._world.current_step

    @property
    def n_steps(self) -> int:
        """Number of steps in a simulation"""
        return self._world.n_steps

    @property
    def default_signing_delay(self) -> int:
        return self._world.default_signing_delay if not self._world.force_signing else 0

    def run_negotiation(
        self,
        issues: Collection[Issue],
        partners: Collection[str | Agent],
        negotiator: Negotiator,
        preferences: Preferences | None = None,
        caller_role: str | None = None,
        roles: Collection[str] | None = None,
        annotation: dict[str, Any] | None = None,
        mechanism_name: str | None = None,
        mechanism_params: dict[str, Any] | None = None,
    ) -> tuple[Contract, NegotiatorMechanismInterface] | None:
        """
        Runs a negotiation until completion

        Args:
            partners: A list of partners to participate in the negotiation.
                      Note that the caller itself may not be in this list which
                      makes it possible for an agent to request a negotaition
                      that it does not participate in. If that is not to be
                      allowed in some world, override this method and explicitly
                      check for these kinds of negotiations and return False.
                      If partners is passed as a single string/`Agent` or as a list
                      containing a single string/`Agent`, then he caller will be added
                      at the beginning of the list. This will only be done if
                      `roles` was passed as None.
            negotiator: The negotiator to be used in the negotiation
            preferences: The preferences. Only needed if the negotiator does not already know it
            caller_role: The role of the caller in the negotiation
            issues: Negotiation issues
            annotation: Extra information to be passed to the `partners` when asking them to join the negotiation
            partners: A list of partners to participate in the negotiation
            roles: The roles of different partners. If None then each role for each partner will be None
            mechanism_name: Name of the mechanism to use. It must be one of the mechanism_names that are supported by the
            `World` or None which means that the `World` should select the mechanism. If None, then `roles` and `my_role`
            must also be None
            mechanism_params: A dict of parameters used to initialize the mechanism object

        Returns:

            A Tuple of a contract and the nmi of the mechanism used to get it in case of success. None otherwise

        """
        a, b = self._world.run_negotiation(
            caller=self.agent,
            issues=issues,
            partners=partners,
            annotation=annotation,
            roles=roles,
            mechanism_name=mechanism_name,
            mechanism_params=mechanism_params,
            negotiator=negotiator,
            preferences=preferences,
            caller_role=caller_role,
        )
        if a is None or b is None:
            return None
        return a, b

    def run_negotiations(
        self,
        issues: list[Issue] | list[list[Issue]],
        partners: list[list[str | Agent]],
        negotiators: list[Negotiator],
        preferences: list[Preferences] = None,
        caller_roles: list[str] = None,
        roles: list[list[str] | None] | None = None,
        annotations: list[dict[str, Any] | None] | None = None,
        mechanism_names: str | list[str] | None = None,
        mechanism_params: dict[str, Any] | list[dict[str, Any]] | None = None,
        all_or_none: bool = False,
    ) -> list[tuple[Contract, NegotiatorMechanismInterface]]:
        """
        Requests to run a set of negotiations simultaneously. Returns after all negotiations are run to completion

        Args:
            partners: A list of partners to participate in the negotiation.
                      Note that the caller itself may not be in this list which
                      makes it possible for an agent to request a negotaition
                      that it does not participate in. If that is not to be
                      allowed in some world, override this method and explicitly
                      check for these kinds of negotiations and return False.
                      If partners is passed as a single string/`Agent` or as a list
                      containing a single string/`Agent`, then he caller will be added
                      at the beginning of the list. This will only be done if
                      `roles` was passed as None.
            issues: Negotiation issues
            negotiators: The negotiator to be used in the negotiation
            ufuns: The utility function. Only needed if the negotiator does not already know it
            caller_roles: The role of the caller in the negotiation
            annotations: Extra information to be passed to the `partners` when asking them to join the negotiation
            partners: A list of partners to participate in the negotiation
            roles: The roles of different partners. If None then each role for each partner will be None
            mechanism_names: Name of the mechanism to use. It must be one of the mechanism_names that are supported by the
            `World` or None which means that the `World` should select the mechanism. If None, then `roles` and `my_role`
            must also be None
            mechanism_params: A dict of parameters used to initialize the mechanism object
            all_or_none: If true, either no negotiations will be started execpt if all partners accepted

        Returns:

             A list of tuples each with two values: contract (None for failure) and nmi (The mechanism info [None if the
             corresponding partner refused to negotiation])

        """
        return self._world.run_negotiations(
            caller=self.agent,
            issues=issues,
            roles=roles,
            annotations=annotations,
            mechanism_names=mechanism_names,
            mechanism_params=mechanism_params,
            partners=partners,
            negotiators=negotiators,
            caller_roles=caller_roles,
            preferences=preferences,
            all_or_none=all_or_none,
        )

    def request_negotiation_about(
        self,
        issues: list[Issue],
        partners: list[str],
        req_id: str,
        roles: list[str] | None = None,
        annotation: dict[str, Any] | None = None,
        mechanism_name: str | None = None,
        mechanism_params: dict[str, Any] | None = None,
        group: str | None = None,
    ) -> bool:
        """
        Requests to start a negotiation with some other agents

        Args:
            req_id:
            issues: Negotiation issues
            annotation: Extra information to be passed to the `partners` when asking them to join the negotiation
            partners: A list of partners to participate in the negotiation.
                      Note that the caller itself may not be in this list which
                      makes it possible for an agent to request a negotaition
                      that it does not participate in. If that is not to be
                      allowed in some world, override this method and explicitly
                      check for these kinds of negotiations and return False.
                      If partners is passed as a single string/`Agent` or as a list
                      containing a single string/`Agent`, then he caller will be added
                      at the beginning of the list. This will only be done if
                      `roles` was passed as None.
            roles: The roles of different partners. If None then each role for each partner will be None
            mechanism_name: Name of the mechanism to use. It must be one of the mechanism_names that are supported by the
            `World` or None which means that the `World` should select the mechanism. If None, then `roles` and `my_role`
            must also be None
            mechanism_params: A dict of parameters used to initialize the mechanism object
            group: An opational identifier for the group to which this negotiation belongs. It is not used by the system
                   but is logged for debugging purposes. Moreover, the agent have access to it through its `negotiations`
                   property.

        Returns:

            List["Agent"] the list of partners who rejected the negotiation if any. If None then the negotiation was
            accepted. If empty then the negotiation was not started from the world manager


        Remarks:

            - The function will create a request ID that will be used in callbacks `on_neg_request_accepted` and
            `on_neg_request_rejected`


        """
        partner_agents = [
            self._world.agents[_] if isinstance(_, str) else _ for _ in partners
        ]
        return self._world.request_negotiation_about(
            req_id=req_id,
            caller=self.agent,
            partners=partner_agents,
            roles=roles,
            issues=issues,
            group=group,
            annotation=annotation,
            mechanism_name=mechanism_name,
            mechanism_params=mechanism_params,
        )

    @property
    def params(self) -> dict[str, Any]:
        """Returns the basic parameters of the world"""
        return self._world.params

    def loginfo(self, msg: str) -> None:
        """
        Logs an INFO message

        Args:
            msg: The message to log

        Returns:

        """
        self._world.loginfo(msg)

    def logwarning(self, msg: str) -> None:
        """
        Logs a WARNING message

        Args:
            msg: The message to log

        Returns:

        """
        self._world.logwarning(msg)

    def logdebug(self, msg: str) -> None:
        """
        Logs a WARNING message

        Args:
            msg: The message to log

        Returns:

        """
        self._world.logdebug(msg)

    def logerror(self, msg: str) -> None:
        """
        Logs a WARNING message

        Args:
            msg: The message to log

        Returns:

        """
        self._world.logerror(msg)

    def loginfo_agent(self, msg: str) -> None:
        """
        Logs an INFO message to the agent's log

        Args:
            msg: The message to log

        Returns:

        """
        self._world.loginfo_agent(self.agent.id, msg)

    def logwarning_agent(self, msg: str) -> None:
        """
        Logs a WARNING message to the agent's log

        Args:
            msg: The message to log

        Returns:

        """
        self._world.logwarning_agent(self.agent.id, msg)

    def logdebug_agent(self, msg: str) -> None:
        """
        Logs a WARNING message to the agent's log

        Args:
            msg: The message to log

        Returns:

        """
        self._world.logdebug_agent(self.agent.id, msg)

    def logerror_agent(self, msg: str) -> None:
        """
        Logs a WARNING message to the agent's log

        Args:
            msg: The message to log

        Returns:

        """
        self._world.logerror_agent(self.agent.id, msg)

    def bb_query(
        self, section: str | list[str] | None, query: Any, query_keys=False
    ) -> dict[str, Any] | None:
        """
        Returns all records in the given section/sections of the bulletin-board that satisfy the query

        Args:
            section: Either a section name, a list of sections or None specifying ALL public sections (see remarks)
            query: The query which is USUALLY a dict with conditions on it when querying values and a RegExp when
            querying keys
            query_keys: Whether the query is to be applied to the keys or values.

        Returns:

            - A dictionary with key:value pairs giving all records that satisfied the given requirements.

        Remarks:

            - A public section is a section with a name that does not start with an underscore
            - If a set of sections is given, and two records in different sections had the same key, only one of them
              will be returned
            - Key queries use regular expressions and match from the beginning using the standard re.match function

        """
        if not self._world.bulletin_board:
            return None
        return self._world.bulletin_board.query(
            section=section, query=query, query_keys=query_keys
        )

    def bb_read(self, section: str, key: str) -> Any | None:
        """
        Reads the value associated with given key from the bulletin board

        Args:
            section: section name
            key: key

        Returns:

            Content of that key in the bulletin-board

        """
        if not self._world.bulletin_board:
            return None
        return self._world.bulletin_board.read(section=section, key=key)

    def bb_record(self, section: str, value: Any, key: str | None = None) -> None:
        """
        Records data in the given section of the bulletin board

        Args:
            section: section name (can contain subsections separated by '/')
            key: The key
            value: The value

        """
        if not self._world.bulletin_board:
            return None
        return self._world.bulletin_board.record(section=section, value=value, key=key)

    def bb_remove(
        self,
        section: list[str] | str | None,
        *,
        query: Any | None = None,
        key: str = None,
        query_keys: bool = False,
        value: Any = None,
    ) -> bool:
        """
        Removes a value or a set of values from the bulletin Board

        Args:
            section: The section
            query: the query to use to select what to remove
            key: the key to remove (no need to give a full query)
            query_keys: Whether to apply the query (if given) to keys or values
            value: Value to be removed

        Returns:
            bool: Success of failure
        """
        if not self._world.bulletin_board:
            return False
        return self._world.bulletin_board.remove(
            section=section, query=query, key=key, query_keys=query_keys, value=value
        )

    @property
    def settings(self):
        if not self._world.bulletin_board:
            return None
        return self._world.bulletin_board.data.get("settings", dict())

    @property
    def initialized(self) -> bool:
        """Was the agent initialized (i.e. was init_() called)"""
        return self.agent._initialized

    @property
    def unsigned_contracts(self) -> list[Contract]:
        """
        All contracts that are not yet signed.
        """
        return list(self.agent._unsigned_contracts)

    @property
    def requested_negotiations(self) -> list[NegotiationRequestInfo]:
        """The negotiations currently requested by the agent.

        Returns:

            A list of negotiation request information objects (`NegotiationRequestInfo`)
        """
        return list(self.agent._requested_negotiations.values())

    @property
    def accepted_negotiation_requests(self) -> list[NegotiationRequestInfo]:
        """A list of negotiation requests sent to this agent that are already accepted by it.

        Remarks:
            - These negotiations did not start yet as they are still not accepted  by all partners.
              Once that happens, they will be moved to `running_negotiations`
        """
        return list(self.agent._accepted_requests.values())

    @property
    def negotiation_requests(self) -> list[NegotiationRequestInfo]:
        """A list of the negotiation requests sent by this agent that are not yet accepted or rejected.

        Remarks:
            - These negotiations did not start yet as they are still not accepted  by all partners.
              Once that happens, they will be moved to `running_negotiations`
        """
        return list(self.agent._requested_negotiations.values())

    @property
    def running_negotiations(self) -> list[RunningNegotiationInfo]:
        """The negotiations currently requested by the agent.

        Returns:

            A list of negotiation information objects (`RunningNegotiationInfo`)
        """
        return list(self.agent._running_negotiations.values())



from __future__ import annotations

import functools
import random
import time
from heapq import heapify, heappop
from math import sqrt
from typing import Callable

import numpy as np
import scipy.optimize as opt

from ..common import MechanismState, Value
from ..models.acceptance import AdaptiveDiscreteAcceptanceModel
from ..negotiators.helpers import PolyAspiration
from ..outcomes import Outcome
from ..sao import AspirationNegotiator, SAONegotiator
from .base import BaseElicitor
from .common import _loc, _scale
from .expectors import (
    AspiringExpector,
    BalancedExpector,
    MaxExpector,
    MeanExpector,
    MinExpector,
)

__all__ = [
    "BasePandoraElicitor",
    "PandoraElicitor",
    "OptimalIncrementalElicitor",
    "FullElicitor",
    "RandomElicitor",
    "weitzman_index_uniform",
    "FastElicitor",
    "MeanElicitor",
    "BalancedElicitor",
    "AspiringElicitor",
    "PessimisticElicitor",
    "OptimisticElicitor",
]


def weitzman_index_uniform(
    loc: float, scale: float, cost: float, time_discount: float = 1.0
) -> float:
    """Implements Weitzman's 1979 Bandora's Box index calculation.

    Args:
        loc: Loc of the uniform distribution to update
        scale: Scale of the uniform distribution to update
        cost: cost of opening a box
        time_discount: time discount. Assumed unity (no discounting) for
                       negotiation.

    Returns:
        The index value of this distribution.

    """
    # assume zi < l
    end = loc + scale
    z = time_discount * (loc + end) / 2.0 - cost

    b = -2 * (end + scale * (1.0 - time_discount) / time_discount)
    c = end * end - 2 * scale * cost / time_discount

    d = b * b - 4 * c
    if d < 0:
        z1 = z2 = -1.0
    else:
        d = sqrt(d)
        z1 = (d - b) / 2.0
        z2 = -(d + b) / 2.0

    if z <= loc and not loc < z1 <= end and not loc < z2 <= end:
        return z
    if z > loc and loc < z1 <= end and not loc < z2 <= end:
        return z1
    if z > loc and not loc < z1 <= end and loc < z2 <= end:
        return z2

    if z <= loc or (z - loc) < 1e-5:
        return z
    elif loc < z1 <= end:
        return z1
    elif loc < z2 <= end:
        return z2
    for _ in (z1, z2):
        if abs(_ - loc) < 1e-5 or abs(_ - end) < 1e-3:
            return _
    print(
        "No solutions are found for (l={}, s={}, c={}, time_discount={}) [{}, {}, {}]".format(
            loc, scale, cost, time_discount, z, z1, z2
        )
    )
    return 0.0


class BasePandoraElicitor(BaseElicitor):
    """
    The base class of all Pandora's box based algorithms.

    Args:
        user: A `User` object that can be elicited
        strategy: An elicitation strategy that determines the questions to
                  be asked and their order.
        base_negotiator: The base negotiator used for responding and proposing
        deep_elicitation: If True, a single elicitation act (`elicit_single`)
                          will elicit the utility value of an outcome until
                          its uncertainty is under `epsilon`.
        opponent_model_factory: A callable to create opponent models. An opponent model is
                                used to estimate acceptance probability of outcomes.
        expector_factory: A callable to create an `Expector`
        single_elicitation_per_round: If True, only a single elicitation act (call to
                                      `elicit_single` is allowed per negotiation round.
        continue_eliciting_past_reserved_val: If True, continue elicitation even if the
                                              estimated utility is under the reserved value.
        epsilon: A small number to represent small uncertainty after which no more elicitation
                 is done.
        true_utility_on_zero_cost: If `True`, zero cost will force the final elicited
                                   value of any outcome to exactly match the utility
                                   function. If `False`, the final utility value after
                                   elicitation may be within `epsilon` from the true
                                   value.
        assume_uniform: If True, assume that all utility distributions are uniform
        user_model_in_index: Use the user model in the index
        precalculated_index: The index is calculated once and never updated
                             when the opponent model probabilities of acceptance
                             change.
        incremental: Only valid if `precalculated_index` is not set. If set,
                     only the outcomes for which the opponemt model acceptance
                     probability have changed will be updated whenever such
                     a change occures. Otherwise, all outcomes will be updated
                     on every change.
        max_aspiration: Maximum aspiration.
        aspiration_type: Aspiration type. Can be "boulware", "linear", "conceder" or a numric
                         exponent.
    """

    def __init__(
        self,
        user: User,
        strategy: EStrategy,
        *,
        base_negotiator: SAONegotiator = AspirationNegotiator(),
        deep_elicitation: bool,
        opponent_model_factory: None
        | (
            Callable[[NegotiatorMechanismInterface], DiscreteAcceptanceModel]
        ) = lambda x: AdaptiveDiscreteAcceptanceModel.from_negotiation(nmi=x),
        expector_factory: Expector | Callable[[], Expector] = MeanExpector,
        single_elicitation_per_round=False,
        continue_eliciting_past_reserved_val=False,
        epsilon=0.001,
        true_utility_on_zero_cost=False,
        assume_uniform=True,
        user_model_in_index=True,
        precalculated_index=False,
        incremental=True,
        max_aspiration=0.99,
        aspiration_type="boulware",
    ) -> None:
        super().__init__(
            strategy=strategy,
            user=user,
            opponent_model_factory=opponent_model_factory,
            expector_factory=expector_factory,
            single_elicitation_per_round=single_elicitation_per_round,
            continue_eliciting_past_reserved_val=continue_eliciting_past_reserved_val,
            epsilon=epsilon,
            true_utility_on_zero_cost=true_utility_on_zero_cost,
            base_negotiator=base_negotiator,
        )
        self.add_capabilities(
            {
                "propose": True,
                "respond": True,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )
        self.my_last_proposals: Outcome | None = None
        self.deep_elicitation = deep_elicitation
        self.elicitation_history = []
        self.cutoff_utility = None
        self.opponent_model = None
        self._elicitation_time = None
        self.offerable_outcomes = (
            []
        )  # will contain outcomes with known or at least elicited utilities
        self.cutoff_utility = None
        self.unknown = None
        self.assume_uniform = assume_uniform
        self.user_model_in_index = user_model_in_index
        self.precalculated_index = precalculated_index
        self.incremental = incremental
        self.__asp = PolyAspiration(max_aspiration, aspiration_type)

    def utility_at(self, x):
        return self.__asp.utility_at(x)

    def utility_on_rejection(self, outcome: Outcome, state: MechanismState) -> Value:
        """Uses the aspiration level as the utility of rejection.

        Remarks:
            - This does not depend on using `AspirationNegotiator` as the
              base_negotiator.
        """
        return self.utility_at(state.relative_time)

    def update_cutoff_utility(self) -> None:
        r"""
        Updates the cutoff utility under which no elicitation is done.

        Remarks:
            - By default it uses :math:`max_{o \in \Omega \cup \phi}{u(o)}` for
              all outcomes :math:`o \in \Omega` and the None outcome representing
              reserved value
        """
        self.cutoff_utility = self.reserved_value
        expected_utilities = [
            self.user_preferences(outcome) for outcome in self.offerable_outcomes
        ]
        if len(expected_utilities) > 0:
            self.cutoff_utility = max(expected_utilities)

    def do_elicit(self, outcome: Outcome, state: MechanismState) -> Value:
        """
        Does a real elicitation step.

        Args:
            outcome: The outcome to elicit
            state: The state at which elicitation is happening

        Remarks:
            - If `deep_elicitation` is set, the strategy is applied until the
              uncertainty in the utility value for `outcome` is less than the
              accuracty limit otherwise, apply it once.
        """
        if not self.deep_elicitation:
            return self.strategy.apply(user=self.user, outcome=outcome)[0]

        while True:
            u, _ = self.strategy.apply(user=self.user, outcome=outcome)
            if isinstance(u, float):
                break
            if u.scale < self.acc_limit:
                u = float(u)
                break

        # we do that after the normal elicitation so that the elicitation time
        # is recorded correctly
        # TODO find a way to avoid calling deep elicitation in this case
        # for speedup. The problem is that I need to calcualate the elicitation
        # time correctly (at least estimate it)
        if self.user.cost_of_asking() == 0.0 and self.true_utility_on_zero_cost:
            return self.user.ufun(outcome)
        return u

    def z_index(self, updated_outcomes: list[Outcome] | None = None):
        """
        Update the internal z-index or create it if needed.

        Args:
            updated_outcomes: A list of the outcomes with updated utility values.
        """
        n_outcomes = self._nmi.n_outcomes
        outcomes = self._nmi.outcomes
        unknown = self.unknown
        if unknown is None:
            unknown = [(-self.reserved_value, _) for _ in range(n_outcomes)]
        else:
            unknown = [_ for _ in unknown if _[1] is not None]

        xw = list(self.preferences.distributions.values())
        if len(unknown) == 0:
            return
        unknown_indices = [_[1] for _ in unknown if _[1] is not None]
        if updated_outcomes is None:
            updated_outcomes = unknown_indices
        else:
            updated_outcomes = [self.indices[_] for _ in updated_outcomes]
            updated_outcomes = list(set(unknown_indices).intersection(updated_outcomes))
        if len(updated_outcomes) == 0:
            return unknown
        z = unknown
        if self.assume_uniform:
            for j, (u, i) in enumerate(unknown):
                if i is None or i not in updated_outcomes:
                    continue
                loc = xw[i].loc if not isinstance(xw[i], float) else xw[i]
                scale = xw[i].scale if not isinstance(xw[i], float) else 0.0
                if self.user_model_in_index:
                    p = self.opponent_model.probability_of_acceptance(outcomes[i])
                    current_loc = loc
                    loc = p * loc + (1 - p) * self.reserved_value
                    scale = (
                        p * (current_loc + scale) + (1 - p) * self.reserved_value - loc
                    )
                cost = self.user.cost_of_asking()
                z[j] = (-weitzman_index_uniform(loc, scale, cost=cost), i)
        else:

            def qualityfun(z, distribution, cost):
                c_estimate = distribution.expect(lambda x: x - z, lb=z, ub=1.0)
                if self.user_model_in_index:
                    p = self.opponent_model.probability_of_acceptance(outcomes[i])
                    c_estimate = p * c_estimate + (1 - p) * self.reserved_value
                return sqrt(c_estimate - cost)

            for j, (u, i) in enumerate(unknown):
                if i is None or i not in updated_outcomes:
                    continue
                cost = self.user.cost_of_asking()
                f = functools.partial(qualityfun, distribution=xw[i], cost=cost)
                z[j] = (
                    -opt.minimize(
                        f, x0=np.asarray([u]), bounds=[(0.0, 1.0)], method="L-BFGS-B"
                    ).x[0],
                    i,
                )
                # we always push the reserved value for the outcome None representing breaking
        heapify(z)
        return z

    def init_unknowns(self):
        """
        Initializes the unknowns list which is a list of Tuples [-u(o), o] for o in outcomes.
        """
        self.unknown = self.z_index(updated_outcomes=None)

    def offer_to_elicit(self) -> tuple[float, int | None]:
        """
        Returns the maximum estimaged utility and the corresponding outocme


        Remarks:
            - This method assumes that each element of the unkown list is a
              `tuple` of negative the utility value and the outcome.
            - It also assumest that the first item in the unknown list contains
              the outocme with maximum estimated utility and the negative
              of the corresponding estimated utility.
        """
        if self.unknown is None:
            self.init_unknowns()
        unknowns = self.unknown
        if len(unknowns) > 0:
            return -unknowns[0][0], unknowns[0][1]
        return self.reserved_value, None

    def update_best_offer_utility(self, outcome: Outcome, u: Value):
        """
        Updates the unknown list (and makes sure it is a heap) given the given utility
        value for the given outcome.
        """
        if self.unknown is None:
            self.init_unknowns()
        self.unknown[0] = (
            -weitzman_index_uniform(_loc(u), _scale(u), self.user.cost_of_asking()),
            self.unknown[0][1],
        )
        heapify(self.unknown)

    def remove_best_offer_from_unknown_list(self) -> tuple[float, int]:
        """
        Removes the best offer from the unkonwn list and returns it.
        """
        if self.unknown is None:
            self.init_unknowns()
        return heappop(self.unknown)

    def elicit_single(self, state: MechanismState):
        """
        Does a single elicitation act

        Args:
            state: mechanism state

        Remarks:
            The process goes as follows:

            1. Find the offer to elicit using `offer_to_elicit`.
            2. If the utility of that offer is less than the cutoff utility,
               or the best offer to elicit is `None`, stop eliciting and return.
            3. call `do_elicit` once.
            4. update the distribution of the elicited outcome
            5. add the elicited outcome to the offerable outcomes
            6. If the utility of the best outcome is a number, remove it
               from the unknown outcomes list otherwise jsut update it
            7. Update the cutoff utility and elicitation history.
        """
        z, best_index = self.offer_to_elicit()
        if z < self.cutoff_utility:
            return False
        if best_index is None:
            return self.continue_eliciting_past_reserved_val
        outcome = self._nmi.outcomes[best_index]
        u = self.do_elicit(outcome, None)
        self.preferences.distributions[outcome] = u
        expected_value = self.offering_utility(outcome, state=state)
        # TODO confirm that offerable_outcomes need not be unique or use
        # a set for it.
        self.offerable_outcomes.append(outcome)
        if isinstance(u, float):
            self.remove_best_offer_from_unknown_list()
        else:
            self.update_best_offer_utility(outcome, u)
        self.cutoff_utility = max(
            (self.cutoff_utility, self.expect(expected_value, state=state))
        )
        self.elicitation_history.append((outcome, u, state.step))
        return True

    def init_elicitation(
        self,
        preferences: IPUtilityFunction | Distribution | None,
        **kwargs,
    ):
        """
        Initializes the elicitation process (called only once).

        Remarks:

            - After calling the parent's `init_elicitation`, this method calls
              `init_unknowns` to initialize the unknown list.
        """
        super().init_elicitation(preferences=preferences, **kwargs)
        strt_time = time.perf_counter()
        self.cutoff_utility = self.reserved_value
        self.unknown = None  # needed as init_unknowns uses unknown
        self.init_unknowns()
        self._elicitation_time += time.perf_counter() - strt_time

    def before_eliciting(self):
        """Called before starting elicitation at every negotiation round.

        Remarks:
            - It just updates cutoff utility
        """
        self.update_cutoff_utility()

    def can_elicit(self) -> bool:
        """
        Checks whether there are any unknowns in the unknowns list.
        """
        if self.unknown is None:
            self.init_unknowns()
        return self.unknown and len(self.unknown) != 0

    def on_opponent_model_updated(
        self, outcomes: list[Outcome], old: list[float], new: list[float]
    ) -> None:
        """
        Called when the opponent model is updated.

        Args:
            outcomes: changed outcomes
            old: old probabilities of acceptance
            new: new probabilities of acceptance

        Remarks:
            Updates the unknown list only if precalculated_index was not set.
        """
        if not self.precalculated_index:
            self.unknown = self.z_index(
                updated_outcomes=outcomes if self.incremental else None
            )


class FullElicitor(BasePandoraElicitor):
    """
    Does full deep elicitation in the first call to `elicit`.
    """

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        epsilon=0.001,
        true_utility_on_zero_cost=False,
        base_negotiator: SAONegotiator = AspirationNegotiator(),
        **kwargs,
    ) -> None:
        kwargs["deep_elicitation"] = True
        super().__init__(
            strategy=strategy,
            user=user,
            epsilon=epsilon,
            true_utility_on_zero_cost=true_utility_on_zero_cost,
            base_negotiator=base_negotiator,
            **kwargs,
        )
        self.elicited = {}

    def update_best_offer_utility(self, outcome: Outcome, u: Value):
        pass

    def init_elicitation(
        self,
        preferences: IPUtilityFunction | Distribution | None,
        **kwargs,
    ):
        super().init_elicitation(preferences=preferences)
        strt_time = time.perf_counter()
        self.elicited = False
        self._elicitation_time += time.perf_counter() - strt_time

    def elicit(self, state: MechanismState):
        if not self.elicited:
            outcomes = self._nmi.outcomes
            utilities = [
                self.expect(self.do_elicit(outcome, state=state), state=state)
                for outcome in self._nmi.outcomes
            ]
            self.offerable_outcomes = list(outcomes)
            self.elicitation_history = [zip(outcomes, utilities)]
            self.elicited = True

    def init_unknowns(self) -> list[tuple[float, int]]:
        self.unknown = []


class RandomElicitor(BasePandoraElicitor):
    """
    Uses a random index instead of the optimal z-index used by the Pandora's
    box solution.
    """

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        deep_elicitation=True,
        true_utility_on_zero_cost=False,
        base_negotiator: SAONegotiator = AspirationNegotiator(),
        opponent_model_factory: None
        | (
            Callable[[NegotiatorMechanismInterface], DiscreteAcceptanceModel]
        ) = lambda x: AdaptiveDiscreteAcceptanceModel.from_negotiation(nmi=x),
        single_elicitation_per_round=False,
        **kwargs,
    ) -> None:
        kwargs["epsilon"] = 0.001
        super().__init__(
            strategy=strategy,
            user=user,
            deep_elicitation=deep_elicitation,
            true_utility_on_zero_cost=true_utility_on_zero_cost,
            opponent_model_factory=opponent_model_factory,
            single_elicitation_per_round=single_elicitation_per_round,
            base_negotiator=base_negotiator,
            **kwargs,
        )

    def init_unknowns(self) -> None:
        n = self._nmi.n_outcomes
        z: list[tuple[float, int | None]] = list(
            zip((-random.random() for _ in range(n + 1)), range(n + 1))
        )
        z[-1] = (z[-1][0], None)
        heapify(z)
        self.unknown = z

    def update_best_offer_utility(self, outcome: Outcome, u: Value):
        pass


class PandoraElicitor(BasePandoraElicitor):
    """
    Implements the original [Baarslag and Gerding]_'s Pandora's box based elicitation
    algorithm (when used with the default parameters).


    Args:
        strategy: The elicitation strategy
        user: The user to elicit


    .. [Baarslag and Gerding] Tim Baarslag and Enrico H. Gerding. 2015.
       Optimal incremental preference elicitation during negotiation.
       In Proceedings of the 24th International Conference on Artificial
       Intelligence (IJCAI’15). AAAI Press, 3–9
       (https://dl.acm.org/doi/10.5555/2832249.2832250).
    """

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        **kwargs,
    ) -> None:
        kwargs.update(
            dict(
                base_negotiator=AspirationNegotiator(),
                opponent_model_factory=lambda x: AdaptiveDiscreteAcceptanceModel.from_negotiation(
                    nmi=x
                ),
                expector_factory=MeanExpector,
                deep_elicitation=True,
                single_elicitation_per_round=False,
                continue_eliciting_past_reserved_val=False,
                epsilon=0.001,
                assume_uniform=True,
                user_model_in_index=True,
                precalculated_index=False,
                incremental=True,
                true_utility_on_zero_cost=False,
            )
        )
        super().__init__(
            strategy=strategy,
            user=user,
            **kwargs,
        )


class FastElicitor(PandoraElicitor):
    """
    Same as `PandoraElicitor` but does not use deep elicitation.
    """

    def __init__(self, *args, **kwargs):
        kwargs["deep_elicitation"] = False
        super().__init__(*args, **kwargs)

    def update_best_offer_utility(self, outcome: Outcome, u: Value):
        """We need not do anything here as we will remove the outcome anyway to the known list"""

    def do_elicit(self, outcome: Outcome, state: MechanismState):
        return self.expect(super().do_elicit(outcome, state), state=state)


class OptimalIncrementalElicitor(FastElicitor):
    """
    Same as `FastElicitor` but uses incremental elicitation which simply
    means that it only updates the index for outcomes that are affected
    by changes in the opponent model.
    """

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        **kwargs,
    ) -> None:
        kwargs.update(dict(incremental=True))
        super().__init__(
            strategy=strategy,
            user=user,
            **kwargs,
        )


class MeanElicitor(OptimalIncrementalElicitor):
    """Same as `OptimalIncrementalElicitor` using `MeanExpector` for
    estimating utilities"""

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        **kwargs,
    ) -> None:
        kwargs.update(dict(expector_factory=MeanExpector))
        super().__init__(
            strategy=strategy,
            user=user,
            **kwargs,
        )


class BalancedElicitor(OptimalIncrementalElicitor):
    """Same as `OptimalIncrementalElicitor` using `MeanExpector` for
    estimating utilities"""

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        **kwargs,
    ) -> None:
        kwargs.update(dict(expector_factory=BalancedExpector))
        super().__init__(
            strategy=strategy,
            user=user,
            **kwargs,
        )


class AspiringElicitor(OptimalIncrementalElicitor):
    """
    Same as `OptimalIncrementalElicitor` using aspiration level for
    estimating utilities.
    """

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        *,
        max_aspiration: float = 1.0,
        aspiration_type: float | str = "linear",
        **kwargs,
    ) -> None:
        kwargs.update(
            dict(
                expector_factory=lambda: AspiringExpector(
                    max_aspiration=max_aspiration,
                    aspiration_type=aspiration_type,
                    nmi=self._nmi,
                )
            )
        )
        super().__init__(
            strategy=strategy,
            user=user,
            max_aspiration=max_aspiration,
            aspiration_type=aspiration_type,
            **kwargs,
        )


class PessimisticElicitor(OptimalIncrementalElicitor):
    """Same as `OptimalIncrementalElicitor` using the minimum to estimate
    utilities."""

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        **kwargs,
    ) -> None:
        kwargs.update(dict(expector_factory=MinExpector))
        super().__init__(
            strategy=strategy,
            user=user,
            **kwargs,
        )


class OptimisticElicitor(OptimalIncrementalElicitor):
    """Same as `OptimalIncrementalElicitor` using the maximum to estimate
    utilities."""

    def __init__(
        self,
        strategy: EStrategy,
        user: User,
        **kwargs,
    ) -> None:
        kwargs.update(dict(expector_factory=MaxExpector))
        super().__init__(
            strategy=strategy,
            user=user,
            **kwargs,
        )
